#include "hough_transform.h"

int hough_opencv(const Mat &srcImg, 
        vector<Point> &points1, vector<Point> &points2)
{
    Mat edges, grayImg;
    vector<Vec2f> lines;
    double threshold1 = 50, threshold2 = 200, 
           rhoStep = 1, thetaStep = CV_PI / 180;
    int houghThreshold = 100, kLines, i;
    cvtColor(srcImg, grayImg, CV_RGB2GRAY);
    Canny(grayImg, edges, threshold1, threshold2);
    namedWindow("Canny (OpenCV)");
    imshow("Canny (OpenCV)", edges);
    HoughLines(edges, lines, rhoStep, thetaStep, houghThreshold);
    kLines = lines.size();
    for (i = 0; i < kLines; i++)
    {
        float rho = lines[i][0], theta = lines[i][1];
        Point pt1, pt2;
        double cosTheta = cos(theta), sinTheta = sin(theta);
        double x0 = cosTheta * rho, y0 = sinTheta * rho;
        pt1.x = cvRound(x0 + 1000 * (-sinTheta));
        pt1.y = cvRound(y0 + 1000 * cosTheta);
        pt2.x = cvRound(x0 - 1000 * (-sinTheta));
        pt2.y = cvRound(y0 - 1000 * cosTheta);
        points1.push_back(pt1);
        points2.push_back(pt2);
    }
    grayImg.release();
    edges.release();
    return 0;
}

int hough_ipp(const Mat &srcImg,
        vector<Point> &points1, vector<Point> &points2)
{
    Ipp32f low = 50.0f, high = 100.0f;
    int minNumPoints = 35, maxLineCount = 40, lineCount, bufSize;
    Ipp8u* pGraySrc, *pBinSrc, *buffer;
    IppiSize pGraySize;
    IppStatus error;

    //   
    Mat grayImg(srcImg.size(), CV_8UC1);
    //   IppiSize     
    pGraySize.width = srcImg.size().width;
    pGraySize.height = srcImg.size().height;
    //      
    error = ippiRGBToGray_8u_C3C1R(srcImg.data, srcImg.step1(), 
                grayImg.data, grayImg.step1(), pGraySize);
    //     
    if (error != ippStsNoErr)
    {
        printf("ERROR!!! ippiRGBToGray_8u_C3C1R(...)\n\n");
        return 1;
    }

    //  
    pGraySrc = (Ipp8u *)grayImg.data;
    //   
    Mat binImg(srcImg.size(), CV_8UC1);
    //       
    int vertSize, horzSize, vertSobelStep, horzSobelStep;
    IppiSize vertSobelSize, horzSobelSize;
    vertSobelSize.width = srcImg.size().width;
    vertSobelSize.height = srcImg.size().height;
    horzSobelSize.width = srcImg.size().width;
    horzSobelSize.height = srcImg.size().height;
    Ipp16s *horzSobel, *vertSobel;
    ippiFilterSobelVertGetBufferSize_8u16s_C1R(vertSobelSize, ippMskSize3x3, &vertSize);
    ippiFilterSobelHorizGetBufferSize_8u16s_C1R(horzSobelSize, ippMskSize3x3, &horzSize);
    if (vertSize < horzSize) 
    {
        vertSize = horzSize;
    }
    ippiCannyGetSize(pGraySize, &vertSize);
    horzSobel = ippiMalloc_16s_C1(srcImg.size().width, srcImg.size().height, &horzSobelStep);
    vertSobel = ippiMalloc_16s_C1(srcImg.size().width, srcImg.size().height, &vertSobelStep);
    buffer = ippsMalloc_8u(vertSize);
    error = ippiFilterSobelVertBorder_8u16s_C1R (pGraySrc, grayImg.step1(), 
                vertSobel, vertSobelStep, 
                vertSobelSize, ippMskSize3x3, ippBorderRepl, 0, buffer);
    if (error != ippStsNoErr)
    {
        printf("ERROR!!! ippiFilterSobelVertBorder_8u16s_C1R(...)\n\n");
		return 1;
    }
    error = ippiFilterSobelHorizBorder_8u16s_C1R(pGraySrc, grayImg.step1(), 
                horzSobel, horzSobelStep, 
                horzSobelSize, ippMskSize3x3, ippBorderRepl, 0, buffer);
    if (error != ippStsNoErr)
    {
        printf("ERROR!!! ippiFilterSobelHorizBorder_8u16s_C1R(...)\n\n");
		return 1;
    }
    error = ippiCanny_16s8u_C1R(horzSobel, horzSobelStep, vertSobel, vertSobelStep, 
                binImg.data, binImg.step1(), pGraySize, low, high, buffer);
    if (error != ippStsNoErr)
    {
        printf("ERROR!!! ippiCanny_16s8u_C1R(...)\n\n");
		return 1;
    }
    ippsFree(buffer);
    ippiFree(horzSobel);
    ippiFree(vertSobel);
    namedWindow("Canny (IPP)");
    imshow("Canny (IPP)", binImg);


    //  
    pBinSrc = (Ipp8u *)binImg.data;
    //     
    IppPointPolar delta;
    delta.rho = 1; //   r
    delta.theta = 1.0f * ((float)CV_PI) / 180.0f; //   theta
    IppPointPolar *lines;
    Ipp8u *pBuffer;
    //    
    error = ippiHoughLineGetSize_8u_C1R(pGraySize, delta, maxLineCount, &bufSize);
    if (error != ippStsNoErr)
    {
        printf("ERROR!!!\n\n");
		return 1;
    }
    //        
    pBuffer = ippsMalloc_8u(bufSize);
    lines = (IppPointPolar *)malloc(sizeof(IppPointPolar) * maxLineCount);
    //   
    error = ippiHoughLine_8u32f_C1R(pBinSrc, binImg.step1(), pGraySize, delta,
        minNumPoints, lines, maxLineCount, &lineCount, pBuffer);
    //     
    if (error != ippStsNoErr)
    {
        printf("ERROR!!!\n\n");
		return 1;
    }
    ippsFree(pBuffer);


    //        
    //   ,   
    for (int lineIdx = 0; lineIdx < lineCount; lineIdx++)
    {
        float rho = lines[lineIdx].rho, theta = lines[lineIdx].theta;
        Point pt1, pt2;
        double cosTheta = cos(theta), sinTheta = sin(theta);
        double x0 = rho * cosTheta, y0 = rho * sinTheta;
        pt1.x = cvRound(x0 + 1000*(-sinTheta));
        pt1.y = cvRound(y0 + 1000*(cosTheta));
        pt2.x = cvRound(x0 - 1000*(-sinTheta));
        pt2.y = cvRound(y0 - 1000*(cosTheta));
        points1.push_back(pt1);
        points2.push_back(pt2);
    }
    binImg.release();
    free(lines);
    return 0;
}

void drawLines(Mat &srcImg, vector<Point> &points1, vector<Point> &points2)
{
    int kLines = points1.size(), i;
    for (i = 0; i < kLines; i++)
    {
        line(srcImg, points1[i], points2[i], Scalar(0, 0, 255), 1);
    }
}
